% conditional_fitting_function:
%
% We perform iterative conditional fitting when learning the covariance
% and variance parameters. This function is the expected conditional 
% likelihood passed to black-box optimizers such as 'fmincon'.
%
% The parameters are organized as follows: there 'maximal_z' different
% instantiations of the joint mixture indicator vector, and 'n' parameters
% per instantiation ('n' regression coefficients plus 1 error variance).
% All 'n' parameters for a particular mixture indicator instantiations are
% added to 'params' before the next batch.

function [f g] = conditional_fitting_function(params)

global data;
global target_subset; global target_var;
global z_post;

target_var_pos = find(target_subset == target_var);
subset = target_subset;
sample_size = size(data, 1);
n = length(subset);
[maximal_z z_universe] = get_z_range(subset);

f = 0;
g = zeros(n * maximal_z, 1);

other_subset = setdiff(subset, target_var);
other_pos = zeros(length(other_subset), 1);
for o = 1:length(other_subset)
    other_pos(o) = find(subset == other_subset(o));
end

for zi = z_universe
    current_z = get_z(subset, zi);    
    if prob_zero(subset, current_z) == 0
       continue
    end
    start = zi;
    %Unpack parameters
    impl_mean = assemble_mean(subset, current_z);
    b = params(((start - 1) * n + 1):(start * n - 1));
    gamma = params(start * n);
    %Data processing
    y_center = data(:, target_var) - impl_mean(target_var_pos);
    mean_others = impl_mean(other_pos)';
    x_center = data(:, other_subset) - mean_others(ones(1, sample_size), :);
    centered = y_center - x_center * b';
    %Function computation
    sq_sum = sum(z_post(:, zi) .* (centered .^ 2));
    z_sum  = sum(z_post(:, zi));
    f      = f - 0.5 * z_sum * log(gamma) - 0.5 * sq_sum / gamma;            
    if ~isempty(gamma(gamma < 1.e-10))
        f = Inf;
        return
    end
    if isnan(f)
        error('NaN here...');
    end
    %Gradient computation
    common_g = z_post(:, zi) .* centered;
    for x = 1:(n - 1)
        g((start - 1) * n + x) = sum(common_g .* x_center(:, x)) / gamma;
    end
    g(start * n) = -0.5 * z_sum / gamma + 0.5 * sq_sum / gamma^2;
end

f = -f / sample_size; % To be used with minimization functions
g = -g / sample_size;
