% create_params:
%
% Allocate and initialize parameter cells. The number of components for
% each mixture indicator (one per variable) is passed as the vector
% 'num_components'.
%
% This initialization needs graph-related information about cliques.
%
% means: means{i} contain the mean parameter for each possible assignment
%        of the clique parents of Y_i;
% vars:  vars{i} contain the variance parameter for each possible assignment
%        of the clique parents of Y_i;
% covs:  covs{i} contain the mean parameter for each possible assignment
%        of the clique intersections of Y_i and Y_j;
% pi:    pi is an length(cliques) x num_components array. Each row
%        specifies the probability distribution of the mixture indicator of
%        the respective clique with respect to the possible 1, 2, ...,
%        'num_components' values;
%
% Author: Ricardo Silva (ricardo@stats.ucl.ac.uk)
% Last modification: Oct 08 2008
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Inputs:
%

function create_params(graph, num_components)

global means; global covs; global vars; global pi_z;

num_vars = length(graph);

means = cell(num_vars, 1);
covs  = cell(num_vars, num_vars);
vars  = cell(num_vars, 1);

% Create distribution for mixture components
pi_z = cell(num_vars, 1);
for i = 1:num_vars
    pi_z{i} = ones(num_components(i), 1) / num_components(i);
end

% Create latent variable model and derive implied covariances and
% variances from it

[cliques, intersections, clique_parents] = build_clique_information(graph);

rand_prop = 5;

coeff = cell(num_vars, 1);
for i = 1:num_vars
    coeff{i} = rand(length(clique_parents{i}), num_components(i)) * rand_prop - 0.5 * rand_prop;
end

for i = 1:num_vars
    %create covariances
    for j = (i + 1):num_vars
        if graph(i, j) ~= 1
           continue;
        end
        covs{i, j} = zeros(num_components(i) * num_components(j), 1);
        total_v = length(covs{i, j});        
        for v = 1:total_v
            values = get_z([i j], v);
            covs{i, j}(v) = 0.;
            for c = 1:length(intersections{i, j})
                cl = intersections{i, j}(c);
                cp1 = find(clique_parents{i} == cl, 1);
                cp2 = find(clique_parents{j} == cl, 1);
                factor = coeff{i}(cp1, values(1)) * coeff{j}(cp2, values(2));
                covs{i, j}(v) = covs{i, j}(v) + factor;
            end            
        end
        covs{j, i} = covs{i, j};
    end
    %create means and variances
    means{i} = zeros(num_components(i), 1);
    vars{i} = zeros(num_components(i), 1);    
    total_v = length(means{i});
    for v = 1:total_v
        vars{i}(v)  = sum(coeff{i}(:, v).^2) + 1;
        means{i}(v) = 2 * rand(1) - 1;
    end
end

