% get_mixture_cov_constraints:
%
% Covariances between some variable 'i' and each element 'j' in
% 'subset' should be sensitive to the level of mixture indicators for
% 'i' and 'j' only. Hence, introduce equality constraints that tie together
% all covariances of 'i' and 'j' in a level of mixture indicator (i, j) 
% to a reference class.
%
% For more information about the format of the output, see the function
% 'get_fixing_cov_constraints'.
%
% Notice that 'i' here is an element of 'subset'.

function [Aeq beq] = get_mixture_cov_constraints(i, subset)

global covs; global vars; global pi_z; global graph;

n         = length(subset);
i_pos     = find(subset == i);
maximal_z = get_z_range(subset);

% Calculate first the number of constraints
num_constraints = 0;
for j_pos = 1:n
    if j_pos == i_pos % variances give quadratic equalities, treated elsewhere
       continue
    end
    j = subset(j_pos);
    for z_i = 1:length(pi_z{i})
        for z_j = 1:length(pi_z{j})
            subsubset = setdiff(subset, [i j]);
            other_maximal_z = get_z_range(subsubset);
            num_constraints = num_constraints + other_maximal_z - 1;
        end
    end
end

if num_constraints > 0
    % there are n parameters per instance of z: n-1 coefficients + variance
    Aeq = zeros(num_constraints, n * maximal_z);
    beq = zeros(num_constraints, 1);
else
    Aeq = [];
    beq = [];
    return
end

% Now, fill in the coefficients of the constraints. First, deal with
% covariance equalities
num_constraints = 0;
for j_pos = setdiff(1:n, i_pos)
    % Covariance equalities (variance equalities are quadratic constraints
    % treated elsewhere: see 'get_mixture_constraints')
    % Notice that we need to allow for 'i' and 'j' to be non-spouses, since
    % we are not applying the Drton-Richardson transformation directly.
    j = subset(j_pos); 
    % Go through all combinations of values for 'i' and 'j'. 
    for z_i = 1:length(pi_z{i})
        for z_j = 1:length(pi_z{j})

            % Get 'z_value', the reference class for the covariance of
            % 'i' and 'j' at mixture level 'z_i', 'z_j' by setting mixture
            % indicators of other variables to assume level 1 
            % (i.e., no need to worry about other cases where the mixture
            % indicators of other variables can take different values,
            % since we know the covariance of 'i', 'j' will be indifferent
            % to that.)

            z_refvalues = ones(n, 1);
            z_refvalues(i_pos) = z_i; z_refvalues(j_pos) = z_j;
            z_refvalue = get_z_inverse(subset, z_refvalues); 

            % Now, iterate through all possible values of the mixture
            % indicators of the other variables (except the reference
            % class).

            subsubset = setdiff(subset, [i j]);
            other_pos = zeros(length(subsubset), 1);
            for o = 1:length(subsubset)
                other_pos(o) = find(subset == subsubset(o));
            end
            [other_maximal_z other_z_universe] = get_z_range(subsubset);
            other_z_universe = setdiff(other_z_universe, 1); %Skip the first instantiation, since this is the reference class.

            for zi_other = other_z_universe
                %Find the corresponding z_index for that
                z_value_other = get_z(subsubset, zi_other);
                z_values = zeros(n, 1);
                z_values(i_pos) = z_i; z_values(j_pos) = z_j;
                z_values(other_pos) = z_value_other;
                z_othervalue = get_z_inverse(subset, z_values);

                % Now, encode this constraint: left-hand side only
                num_constraints = num_constraints + 1;
                for cf = 1:2
                    if cf == 1
                       % Reference class
                       pos = (z_refvalue - 1) * n + 1;
                       values = z_refvalues;
                       sign = 1;
                    else
                       % Other mixture indicator combination
                       pos = (z_othervalue - 1) * n + 1;
                       values = z_values;
                       sign = -1;
                    end
                    for k_pos = 1:n
                        if k_pos == i_pos
                            continue
                        end
                        k = subset(k_pos);                        
                        if j ~= k
                           % Respective covariance of 'j' with 'k'.
                           pk_index = get_parameter_index(j, k, values(j_pos), values(k_pos));
                           if graph(j, k) ~= 0
                              Aeq(num_constraints, pos) = sign * covs{j, k}(pk_index);
                           end
                        else
                           % Respective variance of 'j'
                           Aeq(num_constraints, pos) = sign * vars{j}(values(j_pos));
                        end
                        pos = pos + 1;
                    end
                end                
            end

        end
    end 
end

