% learn:
%
% Main learning procedure. Fits a factorial mixture of Gaussians with
% independent mixture components. Number of levels per mixture component
% is given as an argument to this function. 
%
% Author: Ricardo Silva (ricardo@stats.ucl.ac.uk)
% Last modification: Oct 07 2008
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Inputs:
%
% - latent_init: initialize model with by fitting a latent variable model
% - graph_in: square matrix with graph(i, j) == 1 if and only if the
%             respective vertices are adjacent, 0 otherwise (including
%             diagonal elements graph(i, i));
% - data_in:  matrix with a datapoint per row, fully observed
% - num_components_in: integer indicating how many levels each mixture
%                      component can assume;
% - T:        maximum number of iterations for the learning procedure;
% - T_start:  number of starting points;
% - alpha_in: first  hyperparameter for the inverse gamma expert prior;
% - beta_in:  second hyperparameter for the inverse gamma expert prior;
% - v_b_in:   variance hyperparameter for the Gaussian expert prior;
% - penalize: if true, use penalization method;
%
% Outputs:
%
% - means_out: p x k matrix, p being the number of variables (num_vars) and
%              k the number of values for each mixture indicator 
%              (num_components_in). Provides the mean for each variables, 
%              each mixture component;
% - vars_out:  analogous p x k matrix, but for variances;
% - covs_out:  TODO
% - pi_z_out:  p x k matrix with the distribution for each indicator;
%
% - intercepts_out: intercept parameters for the canonical latent variable
%                   model (CLVM);
% - coeffs_out: coefficient parameters for the CLVM;
% - gammas_out: conditional variance parameters for the CLVM;
% - pi_L_out:   p x k matrix with the distribution for each indicator for
%               the CLVM;

function [means_out, vars_out, covs_out, pi_z_out, ...
          intercepts_out, coeffs_out, gammas_out, pi_L_out, relax_prop] = ...          
         learn(graph_in, data_in, num_components_in, latent_init, ...
               T, T_start, use_relaxed, params_name, load_params, tol)

% Global variable declaration

global data; global graph;
global means; global covs; global vars; global pi_z;
global z_post;

% Initialization

data           = data_in;
graph          = graph_in;
num_components = num_components_in;

disp('*******');
disp('******* FACTORIAL MIXTURE OF GAUSSIANS: maximum likelihood estimator');
disp('******* Version 0.0');
disp('******* Report bugs to Ricardo Silva (ricardo@stats.ucl.ac.uk)');
disp('*******');

% Initialize from latent variable model
if load_params
   load(params_name, '-MAT');
   T_start = 1;
else if latent_init   
        disp('>>> INITIALIZATION: FITTING LATENT VARIABLE MODEL');
        [means, covs, vars, pi_z, intercepts_out, coeffs_out, gammas_out, pi_L_out] = ...
               initialize_from_latents(graph, data, num_components, T, T_start, tol);   
        save(params_name, 'means', 'covs', 'vars', 'pi_z', 'intercepts_out', 'coeffs_out', 'gammas_out', 'pi_L_out');
        disp('>>> END OF INITIALIZATION: start bi-directed model fitting');
        T_start = 1;
    else
        intercepts_out = 0; coeffs_out = 0;
        gammas_out = 0; pi_L_out = 0;
    end
end

% Maximum likelihood: iterations

n          = size(graph, 1);
fixed      = zeros(n, n);
subset     = 1:n; % For future use
best_score = -Inf;
relax_prop = 0; relax_trial = 0;

for random_restart = 1:T_start
 
    if ~latent_init
       create_params(graph, num_components);
    elseif random_restart > 1
       break
    end
    
    z_post = local_expectation(subset);
    score = local_em_score(subset);
    iter_score = score;

    for t = 1:T
        disp(sprintf('** Iteration #%d', t));
        disp('- Expectation...');
        z_post = local_expectation(subset);
        disp(sprintf('Score = %f', score));
        disp('- Maximization...');
        disp(sprintf('> Learning mixture probabilities'));
        learn_pz(subset, fixed);
        score = local_em_score(subset);
        disp(sprintf('>> Score = %f', score));
        disp(sprintf('> Learning means...'));
        for d = 1:n        
            [b gamma] = bartlett_generation(subset(d), subset);
            means{subset(d)} = learn_means(subset(d), subset, b, gamma);
        end
        score = local_em_score(subset);
        disp(sprintf('>> Score = %f', score));
        for d = 1:n
            disp(sprintf('> Learning covariances - Variable: %d', d));
            [b gamma] = bartlett_generation(subset(d), subset);        
            %DEBUG         
               %debug_bartlett(subset(d), subset, b, gamma);
               %debug_score(subset(d), subset);
               %rebuild_from_bartlett(subset(d), subset, b, gamma); 
               %score = local_em_score(subset);
               %disp(sprintf('Score debugging = %f', score));
            %END DEBUG
            if ~isempty(gamma(gamma < 0.0001))
                continue; % Dangerously close to zero: don't do further optimization
            end
            if use_relaxed
               success = learn_covariances_relaxed(subset(d), subset, fixed, b, gamma, score);
               relax_prop  = relax_prop + success;
               relax_trial = relax_trial + 1;
            else
               learn_covariances(subset(d), subset, fixed, b, gamma);
            end
            old_score = score;
            score = local_em_score(subset);
            if score < old_score
                debug_score(subset(d), subset);
                error('Likelihood decrease!');            
            end
            disp(sprintf('Score = %f', score));
        end
        if abs(score - iter_score) < tol
            break
        end
    end

    if score > best_score
        disp(sprintf('****** SCORE IMPROVEMENT: %f --> %f', best_score, score));
        means_best = means;
        vars_best = vars;
        covs_best = covs;
        pi_z_best = pi_z;
        best_score = score;
    end

end

% Wrap up: copy output information and clear memory

relax_prop = relax_prop / relax_trial;

means_out = means_best; vars_out = vars_best; covs_out = covs_best; pi_z_out = pi_z_best;
clear global data; clear global graph;
clear global means; clear global covs; clear global vars; clear global pi_z;
clear global z_post;

disp('Done!');
