
%% learn_mog
%
%Learns a plain mixture of Gaussians model and outputs its predictive 
%log likelihood on a test set
%
% train: training data 
% test: test data
% K: number of mixture components
% T: number of iterations
%
% Last modification: 26/09/2008

function [llik_test llik_train] = learn_mog(train, test, K, T, T_start)

disp('>>>> VANILLA MIXTURE OF GAUSSIANS');
num_train = size(train, 1);
num_vars = size(train, 2);

best_llik_train = -Inf;

for r = 1:T_start
    disp(sprintf('>> RUN #%d', r));
    llik_train = 0;

    means = rand(num_vars, K);
    covs = zeros(num_vars, num_vars, K);
    for k = 1:K
        v = rand(num_vars);
        covs(:, :, k) = v * v' + eye(num_vars) * 0.1;
    end
    post = zeros(num_train, K);
    prob = rand(K, 1);
    prob = prob / sum(prob);

    for t = 1:T

        %Expectation
        for k = 1:K
            inv_k = inv(covs(:, :, k));
            log_detk = log(det(covs(:, :, k)));
            for d = 1:num_train
                centered = train(d, :)' - means(:, k);
                post(d, k) = -0.5 * (log_detk + trace(inv_k * centered * centered')) + log(prob(k));
            end
        end
        for d = 1:num_train
            post(d, :) = post(d, :) - max(post(d, :));
            post(d, :) = exp(post(d, :));
            post(d, :) = post(d, :) / sum(post(d, :));
        end

        %Maximization - covariances
        for k = 1:K
            if (~isempty(find(diag(covs(:, :, k)) < 1.e-10)) || det(covs(:, :, k)) < 1.e-10)
                %disp(sprintf('not updating component %d\n', k));
                continue;
            end
            old_cov = covs(:, :, k);
            covs(:, :, k) = zeros(num_vars, num_vars);
            p_total = 0;
            for d = 1:num_train
                p_total = p_total + post(d, k);
                centered = train(d, :)' - means(:, k);
                covs(:, :, k) = covs(:, :, k) + post(d, k) * centered * centered';
            end
            covs(:, :, k) = covs(:, :, k) / p_total;
            if (~isempty(find(diag(covs(:, :, k)) < 1.e-10)) || det(covs(:, :, k)) < 1.e-10)
                %disp(sprintf('not updating component %d anymore\n', k));
                covs(:, :, k) = old_cov;
                continue;
            end
        end

        %Maximization - means
        for k = 1:K
            means(:, k) = zeros(num_vars, 1);
            p_total = 0;
            for d = 1:num_train
                p_total = p_total + post(d, k);
                means(:, k) = means(:, k) + post(d, k) * train(d, :)';
            end
            means(:, k) = means(:, k) / p_total;
        end

        %Maximization - probs
        for k = 1:K
            prob(k) = sum(post(:, k));
        end
        prob = prob / sum(prob);

        %Calculate log-likelihood
        llik_train = 0;
        for d = 1:num_train
            pred = zeros(K, 1);
            for k = 1:K
                impl_mean = means(:, k);
                impl_cov = covs(:, :, k);
                impl_det = det(impl_cov);
                logdet = log(impl_det);
                inv_impl = inv(impl_cov);    
                dmean = train(d, :)' - impl_mean;
                pred(k) = log(prob(k)) -0.5 * (num_vars * log(2 * pi) + ...
                           logdet + trace(inv_impl *(dmean * dmean')));
            end    
            %max_pred = max(pred);
            %pred = pred - max_pred; 
            llik_train = llik_train + log(sum(exp(pred)));% + K * max_pred;
        end
        llik_train = llik_train / num_train;

        disp(sprintf('Iteration %d, llik_train = %f', t, llik_train));
    end
    
    if llik_train > best_llik_train
       disp(sprintf('****** SCORE IMPROVEMENT: %f --> %f', best_llik_train, llik_train));
       means_best = means;
       covs_best = covs;
       prob_best = prob;
       best_llik_train = llik_train;
    end
end

means = means_best;
covs = covs_best;
prob = prob_best;

%Calculate predictive log-likelihood
num_test = size(test, 1);
llik_test = 0;
for d = 1:num_test
    pred = zeros(K, 1);
    for k = 1:K
        impl_mean = means(:, k);
        impl_cov = covs(:, :, k);
        impl_det = det(impl_cov);
        logdet = log(impl_det);
        inv_impl = inv(impl_cov);    
        dmean = test(d, :)' - impl_mean;
        pred(k) = log(prob(k)) -0.5 * (num_vars * log(2 * pi) + ...
                   logdet + trace(inv_impl *(dmean * dmean')));
    end    
    max_pred = max(pred);
    pred = pred - max_pred; 
    llik_test = llik_test + log(sum(exp(pred))) + max_pred;
end
llik_test = llik_test / num_test;
