% Learn the i-th row of all possible covariance matrices within region
% defined by 'subset'.
%
% Several of the entries are fixed, which will imply linear and
% quadratic constraints on the Bartlett parameters.
%
% IMPORTANT: it is assumed that 'b' is a K x S matrix where 

function [success, new_b, new_gamma] = learn_covariances_relaxed(i, subset, fixed_in, b, gamma, score)
               
global target_var; global target_subset;
global fixed;
global z_post;

fixed = fixed_in;
target_var = i;
target_subset = subset;

[maximal_z z_universe] = get_z_range(subset);

i_pos = find(subset == i);
other_subset = setdiff(subset, i);
other_pos = zeros(length(other_subset), 1);
for v = 1:length(other_pos)
    other_pos(v) = find(subset == other_subset(v));
end

% Common base case: region has only one element
n = length(subset);
if n == 1
   if fixed(1, 1) % i.e., region is a copy of a child, hence no learning
       return;
   end
   new_b = [];
   new_gamma = learn_single_variance(i);
   return
end

% Learn coefficients

b_param = zeros((n - 1) * maximal_z, 1);
for zi = z_universe
    st = (zi - 1) * (n - 1) + 1;
    ed = zi * (n - 1);
    b_param(st:ed) = b(:, zi);
end

new_b_vec = maximize_b_analytical(i, subset, gamma, b_param, fixed);
new_b = zeros(n - 1, maximal_z);
for zi = z_universe
    st = (zi - 1) * (n - 1) + 1;
    ed = zi * (n - 1);
    new_b(:, zi) = new_b_vec(st:ed);
end
rebuild_bartlett_partial(i, subset, new_b, [], 1);

% Now project it back to a starting gamma vector 

if sum(diag(fixed) == true) > 0 
   error('learned_covariances_relaxed: Fixed variances are not allowed!');
end

initial_gamma = get_initial_gamma(i, subset, gamma, new_b);

global ln_gamma_coeff;
global inv_gamma_coeff;
global z_posterior;
global data;

ln_gamma_coeff = zeros(maximal_z, 1);
inv_gamma_coeff = zeros(maximal_z, 1);
sample_size = size(data, 1);

for zi = z_universe
    z_inst = get_z(subset, zi);
    impl_mean = assemble_mean(subset, z_inst);
    ln_gamma_coeff(zi) = sum(z_post(:, zi));

    y_center = data(:, i) - impl_mean(i_pos);
    mean_others = impl_mean(other_pos)';
    x_center = data(:, other_subset) - mean_others(ones(1, sample_size), :);
    centered = y_center - x_center * new_b(:, zi);
    inv_gamma_coeff(zi) = sum(z_posterior(:, zi) .* centered .* centered);
end

global gamma_reference; global offset_matrix; global vars;
[gamma_reference offset_matrix free_index] = build_gamma_mapping(i, subset, new_b);
options = optimset('GradObj','on', 'LargeScale', 'on', 'Display', 'notify', 'DerivativeCheck', 'off');
LB = eps * ones(length(vars{i}), 1); UB = Inf * ones(length(vars{i}), 1);
initial_free_gammas = initial_gamma(free_index);
[out_gamma, fval, exitflag] = fmincon(@neg_gamma_llik_fast, initial_free_gammas, [], [], [], [], LB, UB, [], options);
new_gamma = gamma_reference * out_gamma + offset_matrix;

if sum(new_gamma <= 0)
   error('learn_covariances_relaxed: another gamma problem!');
end
if exitflag < 0
   warning('learn_covariances:var', 'Optimization did not converge!');
end

rebuild_bartlett_partial(i, subset, new_b, new_gamma, 2);

success = true;
old_score = score;
score = local_em_score(subset);
if score < old_score
   disp(sprintf('Relaxation failed'));
   rebuild_from_bartlett(i, subset, b, gamma);
   success = false;
end

clear global target_var; clear global target_subset;
clear global fixed;
